# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from ._conflict import Conflict
from ._doesnotexist import DoesNotExist
from ._fatalexception import FatalException
from ._undecryptable import Undecryptable


__all__: list[str] = [
    'Conflict',
    'ExceptionRaiser',
    'DoesNotExist',
    'FatalException',
    'Undecryptable',
]


class ExceptionRaiser:
    """Mixin class that provides an interface to raise standard exceptions
    defined by the :mod:`libcanonical.types.exceptions` package.
    """
    __module__: str = 'libcanonical.types'
    Conflict        = Conflict
    DoesNotExist    = DoesNotExist
    Undecryptable   = Undecryptable


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Any
from typing import ClassVar
from typing import Literal


LogLevel = Literal[
    #'TRACE',
    'DEBUG',
    'INFO',
    #'SUCCESS',
    'WARNING',
    'ERROR',
    'CRITICAL',
    #'FATAL'
]


class StateLogger:
    __module__: str = 'libcanonical.types'
    logger: ClassVar[logging.Logger]
    logger_name: ClassVar[str] = 'canonical'

    def __init_subclass__(cls, *args: Any, **kwargs: Any) -> None:
        if not hasattr(cls, 'logger'):
            cls.logger = logging.getLogger(cls.__module__)
        if hasattr(cls, 'logger_name'):
            cls.logger = logging.getLogger(__name__)

    def get_logging_parameters(self) -> dict[str, Any]:
        """Hook to return a dictionary holding parameters for the
        local :class:`logging.Logger` instance.
        """
        return {}

    def log(self, level: LogLevel, message: str, *args: Any, **extra: Any):
        match level:
            case 'DEBUG':
                log = self.logger.debug
            case 'INFO':
                log = self.logger.info
            case 'WARNING':
                log = self.logger.warning
            case 'ERROR':
                log = self.logger.error
            case 'CRITICAL':
                log = self.logger.critical
            case _: raise ValueError(f"Unknown log level: {level}")
        return log(
            message,
            *args,
            extra={
                **self.get_logging_parameters(),
                **extra
            }
        )


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
import sys
import pathlib
import tempfile

import pydantic


class EnvironmentBaseModel(pydantic.BaseModel):
    model_config = {'populate_by_name': True}

    loglevel: str = pydantic.Field(
        default='INFO',
        alias='LOGLEVEL'
    )

    tmpdir: pathlib.Path = pydantic.Field(
        default=pathlib.Path(tempfile.gettempdir()),
        alias='TMPDIR'
    )

    vardir: pathlib.Path = pydantic.Field(
        default=pathlib.Path('var').absolute(),
        alias='VARDIR'
    )

    #@pydantic.field_validator('tmpdir', 'vardir', mode='after')
    #@classmethod
    #def validate_directory(
    #    cls,
    #    value: str | pathlib.Path,
    #    writable: bool = True
    #) -> pathlib.Path:
    #    if not os.path.exists(value):
    #        raise ValueError("no such directory")
    #    if not os.access(value, os.R_OK):
    #        raise ValueError("not allowed to read directory")
    #    if not os.path.isdir(value):
    #        raise ValueError("not a directory")
    #    if writable and not os.access(value, os.W_OK):
    #        raise ValueError("directory is not writable")
    #    return pathlib.Path(value)

    @classmethod
    def model_validate_env(cls, environ: dict[str, str] | None = None):
        environ = dict(environ or os.environ)
        try:
            return cls.model_validate(environ)
        except pydantic.ValidationError as exception:
            print("Unable to load environment variables", file=sys.stderr)
            for error in exception.errors():
                print(f"{error['loc'][0]}: {error['msg']} ({repr(error['input'])})")
            raise SystemExit(1)



"""Extensions to the :class:`typer.Typer` command-line application
framework.
"""
import logging.config
from typing import Any

import typer
from typer import Option

from .runtime import LOGGING_CONFIG


__all__: list[str] = [
    'Option',
    'Typer'
]


class Typer(typer.Typer):
    logging_config: dict[str, Any] = LOGGING_CONFIG

    def configure(self):
        pass

    def _configure(self):
        logging.config.dictConfig(self.get_logging_config())
        self.configure()

    def get_logging_config(self):
        return self.logging_config

    def __call__(self, *args: Any, **kwargs: Any):
        self._configure()
        return super().__call__(*args, **kwargs)


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from ._externalstate import ExternalState


class PollingExternalState(ExternalState):
    """An :class:`~libcanonical.bases.ExternalState` implementation that
    uses polling as it's transport mechanism.
    """
    __module__: str = 'libcanonical.bases'

    #: The interval between queries at the remote source for the
    #: current state.
    interval: float = 10.0

    #

    async def poll(self) -> None:
        """Polls the remote source for data."""
        raise NotImplementedError


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import sys


class FatalException(SystemExit):
    """Represent an exception condition which prevents an process from
    continuing to run.
    """
    exitcode: int = 1

    def __init__(self, reason: str):
        sys.stderr.write(reason + '\n')
        sys.stderr.flush()
        super().__init__(self.exitcode)


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import atexit
import asyncio
import threading
import time

from ._lifecycled import Lifecycled
from ._statelogger import StateLogger


class ExternalState(threading.Thread, StateLogger, Lifecycled):
    """The base class for external state implementations. This class should typically not
    be used; instead use either :class:`libcanonical.bases.PollingExternalState` or
    :class:`libcanonical.bases.StreamingExternalState`.
    """
    __module__: str = 'libcanonical.bases'
    _interval: float = 10.0
    last_update: float = 0.0
    must_stop: bool = False
    step: int = 0

    @property
    def age(self):
        return (time.time() - self.last_update)

    def __init__(
        self,
        interval: float = 10.0,
        immediate: bool = False,
        max_age: float | None = None
    ):
        super().__init__(
            daemon=True,
            target=self.__main_event_loop
        )
        self._interval = interval
        self._max_age = max_age
        self._ready_event = threading.Event()
        if immediate:
            self.start()
        atexit.register(self.on_exit)

    def __main_event_loop(self):
        try:
            self.loop = asyncio.get_event_loop()
        except RuntimeError:
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)
        self.loop.run_until_complete(self.setup())
        while True:
            self.step += 1
            if self.must_stop:
                break
            if self.must_update():
                try:
                    self.loop.run_until_complete(self.main_event())
                    self.last_update = time.time()
                    self.setready()
                except Exception as e:
                    self.logger.exception(
                        "Caught fatal %s: %s",
                        type(e).__name__,
                        repr(e)
                    )
            time.sleep(0.1)

        self.loop.run_until_complete(self.teardown())

    async def main_event(self) -> None:
        raise NotImplementedError

    def must_update(self):
        return self.step == 1 or (time.time() - self.last_update) > self._interval

    def on_exit(self):
        self.must_stop = True

    def setready(self):
        self._ready_event.set()

    def wait(self, timeout: float | None = None):
        return self._ready_event.wait(timeout=timeout)

    async def setup(self, reloading: bool = False) -> None:
        raise NotImplementedError

    async def teardown(self, exception: BaseException | None = None) -> bool:
        raise NotImplementedError


# Copyright (C) 2023-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from ._environmentbasemodel import EnvironmentBaseModel
from ._pollingexternalstate import PollingExternalState
from ._regexclassifier import RegexClassifier
from ._statelogger import StateLogger
from ._taggable import Taggable


__all__: list[str] = [
    'EnvironmentBaseModel',
    'PollingExternalState',
    'RegexClassifier',
    'StateLogger',
    'Taggable'
]


# Copyright (C) 2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import functools
import inspect
import logging
import threading
import time
from typing import Callable
from typing import Iterable
from typing import ParamSpec
from typing import TypeVar


P = ParamSpec('P')
T = TypeVar('T')


class _Retry:
    logger: logging.Logger = logging.getLogger('canonical')
    exception_classes: tuple[type[BaseException], ...]
    logging_message: str = "Caught retryable %(exception_class)s (thread: %(thread)s, attempts: %(attempts)s)"

    def __init__(
        self,
        types: Iterable[type[BaseException]],
        max_attempts: int | None = None,
        delay: int | float = 10,
        reason: str | None = None
    ):
        self.delay = delay
        self.exception_classes = tuple(types)
        self.max_attempts = max_attempts
        if reason:
            self.logging_message = reason

    def on_exception(self, exception: BaseException, attempts: int):
        if not isinstance(exception, self.exception_classes):
            return True
        if self.max_attempts and attempts > self.max_attempts:
            return True
        t = threading.current_thread()
        self.logger.warning(
            self.logging_message,
            extra={
                'attempts': attempts,
                'exception_class': f'{type(exception).__name__}',
                'thread_id': t.ident
            }
        )

    def call_sync(self, f: Callable[P, T]) -> Callable[P, T]:
        handles = self.exception_classes

        @functools.wraps(f)
        def d(*args: P.args, **kwargs: P.kwargs):
            attempts = 0
            while True:
                attempts += 1
                try:
                    result = f(*args, **kwargs)
                    break
                except KeyboardInterrupt:
                    raise
                except Exception as e:
                    if not isinstance(e, handles):
                        raise
                    if self.max_attempts and attempts > self.max_attempts:
                        raise e from e
                    t = threading.current_thread()
                    self.logger.warning(
                        self.logging_message,
                        extra={
                            'attempts': attempts,
                            'exception_class': f'{type(e).__name__}',
                            'thread_id': t.ident
                        }
                    )
                    time.sleep(self.delay)
            return result

        return d # type: ignore

    def __call__(
        self,
        f: Callable[P, T]
    ) -> Callable[P, T]:
        if not inspect.iscoroutinefunction(f):
            return self.call_sync(f)
        assert inspect.iscoroutinefunction(f)
        handles = self.exception_classes

        @functools.wraps(f)
        async def d(*args: P.args, **kwargs: P.kwargs):
            attempts = 0
            while True:
                attempts += 1
                try:
                    result = await f(*args, **kwargs)
                    break
                except KeyboardInterrupt:
                    raise
                except Exception as e:
                    if not isinstance(e, handles):
                        raise
                    if self.max_attempts and attempts > self.max_attempts:
                        raise e from e
                    t = threading.current_thread()
                    self.logger.warning(
                        self.logging_message,
                        extra={
                            'attempts': attempts,
                            'exception_class': f'{type(e).__name__}',
                            'thread_id': t.ident
                        }
                    )
                    await asyncio.sleep(self.delay)
            return result

        return d # type: ignore


retry = _Retry


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import enum


class ApplicationRuntimeState(int, enum.Enum):
    BOOTING         = 0
    STARTUP         = 1
    LIVE            = 2
    READY           = 3
    BUSY            = 4
    TEARDOWN        = -1
    BOOTFAILURE     = -2
    RELOADFAILURE   = -3
    FATAL           = -99


# Copyright (C) 2023-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Generic
from typing import TypeVar
from typing import Hashable


T = TypeVar('T', bound=Hashable)


class Taggable(Generic[T]):
    """A base class that provides tagging functionality."""
    tags: set[T]

    def __init__(self, *args: Any, **kwargs: Any):
        self.tags = set()

    def tag(self, tag: T) -> bool:
        added = tag not in self.tags
        self.tags.add(tag)
        return added

    def is_tagged(self, tag: T):
        return tag in self.tags

    def untag(self, tag: T) -> bool:
        removed = tag in self.tags
        self.tags.remove(tag)
        return removed


# Copyright (C) 2023-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import TypeVar

from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic.json_schema import JsonSchemaValue
from pydantic import GetJsonSchemaHandler


__all__: list[str] = [
    'HexEncoded'
]

T = TypeVar('T', bound='HexEncoded')


class HexEncoded(bytes):
    __module__: str = 'libcanonical.types'

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.str_schema(),
            python_schema=core_schema.chain_schema([
                core_schema.union_schema([
                    core_schema.is_instance_schema(bytes),
                    core_schema.chain_schema([
                        core_schema.is_instance_schema(str),
                        core_schema.no_info_plain_validator_function(cls.fromstring),
                        core_schema.no_info_plain_validator_function(cls.validate)
                    ])
                ]),
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(cls.serialize)
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.str_schema())

    @classmethod
    def fromstring(cls, v: str):
        if v.startswith('0x'):
            v = v[2:]
        return cls.fromhex(v)

    @classmethod
    def validate(cls, instance: T) -> T:
        return instance

    def serialize(self) -> str:
        return f"0x{self.hex()}"


# Copyright (C) 2023-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
from typing import cast
from typing import Any
from typing import Generic
from typing import TypeVar


T = TypeVar('T')


class RegexClassifier(Generic[T]):
    """Classifies object T with tags."""
    patterns: list[re.Pattern[str] | str] | re.Pattern[str]
    flags: int = 0

    @classmethod
    def __init_subclass__(cls, *args: Any, **kwargs: Any) -> None:
        if not hasattr(cls, 'patterns'):
            return
        if isinstance(cls.patterns, str):
            cls.patterns = re.compile(cls.patterns, flags=cls.flags)
        if isinstance(cls.patterns, re.Pattern):
            cls.patterns = [cls.patterns]
        for i, pattern in enumerate(cls.patterns):
            if isinstance(pattern, re.Pattern):
                continue
            cls.patterns[i] = re.compile(pattern, flags=cls.flags)

    def __init__(self) -> None:
        if not hasattr(self, 'patterns'):
            self.patterns = []

    def add_patterns(self, patterns: list[str | re.Pattern[str]]):
        assert isinstance(self.patterns, list)
        for p in patterns:
            if isinstance(p, str):
                p = re.compile(p, self.flags)
            self.patterns.append(p)

    def wants(self, obj: T) -> bool:
        """Return a boolean indicating if the classifier is interested
        in `obj`.
        """
        assert isinstance(self.patterns, list)
        return any([self.run_pattern(cast(re.Pattern[str], p), obj) for p in self.patterns])

    def run_pattern(self, pattern: re.Pattern[str], obj: T) -> re.Match[str] | None:
        if isinstance(obj, str) or hasattr(obj, '__str__'):
            return pattern.match(str(obj))
        raise TypeError(
            f"{type(obj).__name__} can not be cast to string. Override the "
            f"{type(self).__name__}.run_pattern() method or supply a string "
            "type."
        )

    def classify(self, obj: T) -> None:
        raise NotImplementedError


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
import signal
from typing import Callable
from types import FrameType
from typing import Any


class SignalHandler:
    """Mixin class to bind interrupt signals. An interrupt signal is bound
    by defining an ``on_{lowercased signal}`` on the inherting class.

    .. code:: python

        class Main(SignalHandler):

            def on_sigint(self):
                print("Caught SIGINT!")

    Note that if the inherting class override the :meth:`__init__` method,
    it must call :func:`super()`. :class:`SignalHandler` ignores any
    positional and keyword arguments.
    """

    def __init__(self, *_: Any, **__: Any) -> None:
        self.__bind_signals()

    def __bind_signals(self):
        f: Callable[[FrameType], None] | Callable[[], None] | None
        for _, value in inspect.getmembers(signal):
            if not isinstance(value, signal.Signals):
                continue
            f = getattr(self, f'on_{str.lower(value.name)}', None)
            if f is None:
                continue
            signal.signal(value, self.__signal_handler(f))

    def __signal_handler(
        self,
        f: Callable[[FrameType], None] | Callable[[], None]
    ) -> Callable[[int, FrameType | None], None]:
        handler: Callable[[int, FrameType | None], None]
        signature = inspect.signature(f)
        arglen = len(signature.parameters.values())
        match arglen:
            case 0: handler = lambda _, __: f() # type: ignore
            case 1: handler = lambda _, frame: f(frame) # type: ignore
            case _:
                raise TypeError(
                    "Invalid signature for signal handler "
                    f"{type(self).__name__}.{f.__name__}, "
                    f"expected at most 1 parameters, got {arglen}"
                )
        return handler


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from typing import Any
from typing import Coroutine
from types import TracebackType


class Lifecycled:
    __module__: str = 'libcanonical.bases'

    def setup(
        self,
        reloading: bool = False
    ) -> None | Coroutine[Any, Any, None]:
        raise NotImplementedError

    def teardown(
        self,
        exception: BaseException | None = None
    ) -> bool | Coroutine[Any, Any, bool]:
        raise NotImplementedError

    async def __aenter__(self):
        result = self.setup(reloading=False)
        if inspect.isawaitable(result):
            await result
        return self

    async def __aexit__(
        self,
        cls: type[BaseException] | None = None,
        exception: BaseException | None = None,
        traceback: TracebackType | None = None
    ) -> bool:
        result = self.teardown()
        if inspect.isawaitable(result):
            await result
        assert isinstance(result, bool)
        return result


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import time
from typing import Callable
from typing import TypeVar
from types import TracebackType


R = TypeVar('R')


class DelayContext:

    @property
    def remaining(self):
        elapsed = time.monotonic() - self.started
        remaining = max(self.delay - elapsed, 0.0)
        if remaining > 0.0 and self.timeout:
            timeout = self.timeout - self.latency
            remaining = min(timeout, remaining)
        return remaining

    def __init__(self, delay: float, timeout: float | None = None, latency: float = 1.0):
        self.delay = delay
        self.timeout = timeout or 0.0
        self.latency = latency
        if self.timeout == 0.0:
            self.latency = 0.0

    def sleep(self, func: Callable[[float], R]) -> R:
        return func(self.remaining)

    async def __aenter__(self):
        self.started = time.monotonic()
        return self

    async def __aexit__(
        self,
        cls: type[BaseException],
        exc: BaseException,
        tb: TracebackType
    ):
        if exc:
            raise exc
        await self.sleep(asyncio.sleep)

    def __enter__(self):
        self.started = time.monotonic()
        return self

    def __exit__(
        self,
        cls: type[BaseException],
        exc: BaseException,
        tb: TracebackType
    ):
        if exc:
            raise exc
        self.sleep(time.sleep)


delay = DelayContext